#define _REENTRANT
#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <X11/X.h>
#include <X11/StringDefs.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <thread.h>

/* function prototypes */
int ThrCB();
void *thr_sub(void *);

main(int argc, char **argv)
{
Widget 	toplevel, base, button[200];
Arg 	wargs[10];
int	i, MT=0;
char 	but_label[5];

/* check arguments */
if (argc > 3 || argc < 2) 
	printf("Usage: %s <number of buttons> [MT]\n", argv[0]), exit(0);

/* check multithreaded mode */
if (argc == 3 && strcmp(argv[2], "MT") == 0) MT = 1;
if (atoi(argv[1]) > 200) 
	printf("Too many buttons...\n"), exit(1);

/* set up our main window */
toplevel = XtInitialize(argv[0], "ThreadWin", NULL, 0, &argc, argv);

/* set some arguments for our Row/Column Widget */
XtSetArg(wargs[0], XmNorientation, XmHORIZONTAL);
XtSetArg(wargs[1], XmNentryAlignment, XmALIGNMENT_CENTER);
XtSetArg(wargs[2], XmNisAligned, True);
XtSetArg(wargs[3], XmNnumColumns,(int)floor(sqrt((double)atoi(argv[1]))));
XtSetArg(wargs[4], XmNpacking, XmPACK_COLUMN);

/* create the Row/Column Widget */
base=XtCreateManagedWidget("base", xmRowColumnWidgetClass, toplevel, 
wargs, 5);

/* create the button widgets with the button number as it label */
for (i=0;i<atoi(argv[1]);i++) {
	sprintf(but_label, "%d", i);
	XtSetArg(wargs[0], XmNlabelString, XmStringCreate(but_label, XmSTRING_DEFAULT_CHARSET));
	button[i] = XmCreatePushButton(base, "button", wargs, 1);

	/* tell the button to call the ThrCB() routine when pushed */
	XtAddCallback(button[i], XmNactivateCallback, ThrCB, MT);
	}

/* set the level of concurrency to the number of processors on-line */
thr_setconcurrency(sysconf(_SC_NPROCESSORS_ONLN)+1);

/* manage the buttons and go into the X event loop */
XtManageChildren(button, atoi(argv[1]));
XtRealizeWidget(toplevel);
XtMainLoop();
return(0);
}


/* 
	Button callback routine -- Called when a button is pushed 
*/

ThrCB(Widget w, int MT)
{
/* check to see if we should run multithreaded 
   if so, create a thread that runs the thr_sub() routine
   if not, just call the thr_sub() routine  */

if (MT) thr_create(NULL, 0, thr_sub, (void *)w, THR_DETACHED,  NULL);
else thr_sub((void *)w);
return(0);
}




/* 
	thr_sub() -- called from ThrCB() routine.  This is the
 	routine that changes the button and simulates some work
*/

void *thr_sub(void *arg)
{
Widget w = (Widget) arg;
Pixel fg, bg;
XEvent event;
Arg warg[2];
int j;
Display *disp;
Window win;

/* grey out the button */
XtSetSensitive(w, False);

/* get the button's foreground and background colors */
XtSetArg(warg[0], XmNforeground, &fg);
XtSetArg(warg[1], XmNbackground, &bg);
XtGetValues(w, warg, 2);

/* swap the button's foreground and background colors */
XtSetArg(warg[0], XmNforeground, bg);
XtSetArg(warg[1], XmNbackground, fg);
XtSetValues(w, warg, 2);

/* force a pointer movement event to make the button changes
   visible */

disp = XtDisplay(w);
win = XtWindow(w);
event.xmotion.display = disp;
event.xmotion.window = win;
XSendEvent(disp, win, TRUE, PointerMotionMask, &event);
XFlush(disp);

/* simulate some processing */
for (j=0;j<40000000;j++);

/* swap the button colors back */
XtSetArg(warg[0], XmNforeground, fg);
XtSetArg(warg[1], XmNbackground, bg);
XtSetValues(w, warg, 2);

/* make the button active again */
XtSetSensitive(w, True);

/* force another X event so that the window will redisplay */
XSendEvent(disp, win, TRUE, PointerMotionMask, &event);
XFlush(disp);

return((void *)0);
}
