#define _REEENTRANT
#include <stdio.h>
#include <thread.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/uio.h>

#define BUFSIZE 512
#define BUFCNT  4

/* this is the data structure that is used between the producer
   and consumer threads */

struct {
  	char buffer[BUFCNT][BUFSIZE];
	int byteinbuf[BUFCNT];
	mutex_t buflock;
	mutex_t donelock;
	cond_t adddata;
	cond_t remdata;
	int nextadd, nextrem, occ, done;
} Buf;

/* function prototype */
void *consumer(void *);

main(int argc, char **argv)
{
int ifd, ofd;
thread_t cons_thr;

/* check the command line arguments */
if (argc != 3)
	printf("Usage: %s <infile> <outfile>\n", argv[0]), exit(0);

/* open the input file for the producer to use */
if ((ifd = open(argv[1], O_RDONLY)) == -1)
	{
	fprintf(stderr, "Can't open file %s\n", argv[1]);
	exit(1);
	}

/* open the output file for the consumer to use */
if ((ofd = open(argv[2], O_WRONLY|O_CREAT, 0666)) == -1)
	{
	fprintf(stderr, "Can't open file %s\n", argv[2]);
	exit(1);
	}

/* zero the counters */
Buf.nextadd = Buf.nextrem = Buf.occ = Buf.done = 0;

/* set the thread concurrency to 2 so the producer and consumer can
   run concurrently */

thr_setconcurrency(2);

/* create the consumer thread */
thr_create(NULL, 0, consumer, (void *)ofd, NULL, &cons_thr);

/* the producer ! */
while (1) {

	/* lock the mutex */
	mutex_lock(&Buf.buflock);

	/* check to see if any buffers are empty */
	/* If not then wait for that condition to become true */

	while (Buf.occ == BUFCNT)
		cond_wait(&Buf.remdata, &Buf.buflock);

	/* read from the file and put data into a buffer */
	Buf.byteinbuf[Buf.nextadd] = read(ifd,Buf.buffer[Buf.nextadd],BUFSIZE);

	/* check to see if done reading */
	if (Buf.byteinbuf[Buf.nextadd] == 0) {

  		/* lock the done lock */
		mutex_lock(&Buf.donelock);

		/* set the done flag and release the mutex lock */
		Buf.done = 1;

		mutex_unlock(&Buf.donelock);

		/* signal the consumer to start consuming */
		cond_signal(&Buf.adddata);

		/* release the buffer mutex */
		mutex_unlock(&Buf.buflock);

		/* leave the while looop */
		break;
		}

	/* set the next buffer to fill */
	Buf.nextadd = ++Buf.nextadd % BUFCNT;

	/* increment the number of buffers that are filled */
	Buf.occ++;

	/* signal the consumer to start consuming */
	cond_signal(&Buf.adddata);

	/* release the mutex */
	mutex_unlock(&Buf.buflock);
	}

close(ifd);

/* wait for the consumer to finish */
thr_join(cons_thr, 0, NULL);

/* exit the program */
return(0);
}


/* The consumer thread */
void *consumer(void *arg)
{
int fd = (int) arg;

/* check to see if any buffers are filled or if the done flag is set */
while (1) {

	/* lock the mutex */
	mutex_lock(&Buf.buflock);

 	if (!Buf.occ && Buf.done) {
	   mutex_unlock(&Buf.buflock);
  	   break;
	   }

	/* check to see if any buffers are filled */
	/* if not then wait for the condition to become true */

	while (Buf.occ == 0 && !Buf.done)
		cond_wait(&Buf.adddata, &Buf.buflock);

	/* write the data from the buffer to the file */
	write(fd, Buf.buffer[Buf.nextrem], Buf.byteinbuf[Buf.nextrem]);

	/* set the next buffer to write from */
	Buf.nextrem = ++Buf.nextrem % BUFCNT;

	/* decrement the number of buffers that are full */
	Buf.occ--;

	/* signal the producer that a buffer is empty */
	cond_signal(&Buf.remdata);

	/* release the mutex */
	mutex_unlock(&Buf.buflock);
	}

/* exit the thread */
thr_exit((void *)0);
}
